#include <stdlib.h>
#include <iostream>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include "myrand.h"

typedef uint64_t BB; // Bitboard == 64 unsigned bits

void
print(BB bb)
{
  for(int rank=8;rank--;) {
    for(int file=0;file<8;++file) {
      std::cout<< "01"[1&(bb>>(rank*8+file))];
    }
    std::cout<<std::endl;
  }
}

int
count(BB bb)
{
  int res;
  res=0;
  for(int i=64;i--;)
    if (1&(bb>>i)) ++res;
  if (res<0)
    return res;
  bb = bb-((bb>>1)&0x5555555555555555ll); // 2 bits per sum
  bb = (bb&0x3333333333333333ll)+((bb>>2)&0x3333333333333333ll); // 3 bits
  bb = bb+(bb>>32); // 4 bits per sum
  bb = (bb&0x0f0f0f0f)+((bb>>4)&0x0f0f0f0f); // 5 bits per sum
  bb = bb+(bb>>16); // 6 bits per sum
  return 0xff&(bb+(bb>>8)); // 7 bit sum
}

int
is_white_piece(char c)
{
  if ('R'==c) return 1;
  return 0;
}

int
max(int x, int y)
{
  return x>y?x:y;
}

int
min(int x, int y)
{
  return x<y?x:y;
}

int
abs(int v)
{
  return max(v, -v);
}

int
distance(int sq1, int sq2)
{
  int d1 = abs((7&sq1)-(7&sq2));
  int d2 = abs((070&sq1)-(070&sq2))>>3;
  return max(d1,d2);
}

BB
squares_in_between(int sq1, int sq2)
{
  if (distance(sq1,sq2)<2)
    return (1ll<<sq1)|(1ll<<sq2);
  int sqm = ((7&sq1)+(7&sq2))/2 + (070&(((070&sq1)+(070&sq2))/2));
  return squares_in_between(sq1, sqm)|squares_in_between(sq2, sqm);
}

BB
blocks_attack(char piece, int ksq, int psq)
{
  switch(piece) {
  case 'R':
    int tsq = psq^ksq;
    tsq = tsq|tsq<<1|tsq<<2;
    tsq = tsq&tsq>>3;
    if (!(tsq&1<<2)) { // same rank or file
      return squares_in_between(ksq, psq);
    } else return ~0ll;
  }
  exit(1); // to please compiler
}


BB *
compute_illegal(char *pl)
{
  int count = strlen(pl);
  BB *t = (BB*)malloc(8<<(6*(count+1)));

  for (int ksq=64;ksq--;) { // black king
    BB bb = 1ll<<ksq;
    bb = bb|(0xfefefefefefefefell&(bb<<1))|(0x7f7f7f7f7f7f7f7fll&(bb>>1));
    bb = bb|(bb<<8)|(bb>>8); // squares of or next to black king
    BB *tt = &t[(1<<(6*count))*ksq];
    for (int i=1<<(6*count);i--;) {
      BB bb2=bb;
      for (int k=count;k--;) {
	bb2 |= 1ll<<((i>>(6*k))&077);
      }
      tt[i] = bb2;
    }

    for (int pi=count;pi--;)
      if (is_white_piece(pl[pi])) {
	for (int psq=64;psq--;) {
	  BB block = blocks_attack(pl[pi],ksq,psq);
	  int pp[count];
	  for (int k=count;k--;) pp[k] = k==pi?psq:077;
	  do {
	    int k;
	    for (k=count;k--;) {
	      if (k!=pi && 0!=(block&1ll<<pp[k])) break;
	    }
	    if (k<0) {
	      int m=0;
	      for (k=count;k--;) {
		m<<=6;
		m+=pp[k];
	      }
	      tt[m] |= ~block;
	    }
	    for(k=count;k--;) { // cute loop
	      if (k==pi) continue;
	      if (pp[k]--) break;
	      pp[k]=077;
	    }
	    if(k<0) break;
	  } while(1);
	}
      }
  }
  for(int p1=count+2;p1--;)
    for(int p2=p1;p2--;)
      for (int sq=64;sq--;) {
	int pm, pv;
	BB bb;
	if (p2) {
	  pm = 077<<(6*(p1-1));
	  pm |= 077<<(6*(p2-1));
	  pv = sq<<(6*(p1-1));
	  pv |= sq<<(6*(p2-1));
	  bb = ~0ll;
	} else {
	  pm = 077<<(6*(p1-1));
	  pv = sq<<(6*(p1-1));
	  bb = 1ll<<sq;
	}
	for (int i=1ll<<(6*(count+1));(i=(i-1)&~pm);)
	  t[i|pv] |= bb;
      }
  return t;
}

int
count_bits(BB *t, char *pl)
{
  int total=0;
  for (int i = 1<<(6*(strlen(pl)+1));i--;)
    total += count(t[i]);

  return total;
}

int
main(int argc, char *argv[])
{
  myinit();
  if(argc!=2)
    exit(1);
  if(1)for (int i=10*64;i--;) {
    BB *ill2 = compute_illegal(argv[1]);
    free(ill2);
  }
  BB *ill = compute_illegal(argv[1]);
  int count=strlen(argv[1]);
  if(0&&write(1,ill,1<<(6*count+9)))
    exit(0);
  long long illegal = count_bits(ill, argv[1]);
  long long legal = (1<<(6*(count+2)))-illegal;
  std::cout << legal << " legal positions" << std::endl;
  std::cout << illegal << " illegal positions" << std::endl;
  printf("%o\n", count);
  for (int x=1;x--;) {
    std::cout << "Hello World!" << std::endl;
    int p=myrand()&((64<<(6*count))-1);
    p=07766;
    printf("%o\n", p);
    print(ill[p]);
  }
  exit(0);
}
